##
# Function: obtenerOrdenacion
#
# Author: Daniel Henrndez Lobato
#
# Arguments:
#		C <- correlation matrix as described by Zhou et al.		
#
# returns:
#		Ordered sequence of predictors
#

orderedAggregation <- function(C) {

	order <- NULL

	# We select the best single element

	order <- which.min(as.vector(diag(C)))

	# We iteratively include the best element using the greedy algorithm

	nElements <- 1;

	while (length(order) != nrow(C)) {
		
		if (nElements %% 10 == 0) cat(".")

		contribution <- rep(0, nrow(C) - nElements)
		toInclude <- (1 : nrow(C))[ -order ]
		
		for (i in 1 : length(toInclude))
			contribution[ i ] <- 2 * sum(C[ toInclude[ i ], order ]) + C[ toInclude[ i ], toInclude[ i ] ]

		# We select the best element 

		nElements <- nElements + 1
		order <- c(order, toInclude[ which.min(contribution) ])
	}

	cat("\n")
	
	order
}


