#include <R.h>
#include <Rinternals.h>
#include <R_ext/Rdynload.h>
#include <stdio.h>
#include <math.h>

// SDP library

#include <declarations.h>

int *approx_SDP(double **H, double **X, int dim, int k, double *value);
int solve_SDP(double **H, double **D, int dim, int k);

void R_init_Rinterface(DllInfo *info) {}
void R_unload_Rinterface(DllInfo *info) {}

void selectSDP(double *matrix,  double *sizeBag, double *pruningRate, int *indexes) {
	int *S, *Stmp;
	double vS, vStmp;
	double **X, **H, **D, sum;
	int nClasf, i, j, nsel;

	nClasf = (int) *sizeBag;

	H = (double**) malloc(sizeof(double*) * (nClasf + 1 ));
	for (i = 0; i < (nClasf + 1); i++) H[ i ] = (double *) malloc(sizeof(double) * (nClasf + 1));
	D = (double**) malloc(sizeof(double*) * (nClasf + 1 ));
	for (i = 0; i < (nClasf + 1); i++) D[ i ] = (double *) malloc(sizeof(double) * (nClasf + 1));

	// We compute the diagonal of the G matrix and the off diagonal terms

//	XXX Mistake
/*	for (i = 0; i < nClasf; i++)
		H[ i + 1 ][ i + 1 ] = matrix[ i * nClasf ];
	for (i = 0; i < (nClasf - 1); i++)
		for (j = (i + 1); j < nClasf; j++) {
			H[ i + 1 ][ j + 1 ] = matrix[ i * nClasf + j ];
			H[ j + 1 ][ i + 1 ] = H[ i + 1 ][ j + 1 ];
		}*/

	for (i = 0; i < nClasf; i++)
		for (j = i; j < nClasf; j++) {
			H[ i + 1 ][ j + 1 ] = matrix[ i * nClasf + j ];
			H[ j + 1 ][ i + 1 ] = H[ i + 1 ][ j + 1 ];
		}

	// We compute e~G terms in the H matrix

	for (i = 0; i < nClasf; i++) {
		sum = 0;
		for (j = 0; j < nClasf; j++) sum += H[ i + 1 ][ j + 1 ];
		H[ 0 ][ i + 1 ] = sum;
		H[ i + 1 ][ 0 ] = H[ 0 ][ i + 1 ];
	}

	// We compute e^t~Ge terms in the H matrix

	H[ 0 ][ 0 ] = 0;
	for (i = 0; i < nClasf; i++) H[ 0 ][ 0 ] += H[ 0 ][ i + 1 ];

	// We compute the D matrix

	for (i = 0; i < nClasf + 1; i++) 
		for (j = 0; j < nClasf + 1; j++) 
			D[ i ][ j ] = 0;			

	for (i = 0; i < nClasf; i++) {
		D[ i + 1 ][ i + 1 ] = 1;
		D[ 0 ][ i + 1 ] = 1;
		D[ i + 1 ][ 0 ] = 1;
	}

	D[ 0 ][ 0 ] = nClasf;

	// We obtain the solution of the SDP opt problem called X

	if (solve_SDP(H,D, nClasf + 1, (int) (*pruningRate * nClasf + .5)) != 1) {
		printf("Error while solving SDP problem.\n");
		exit(0);
	}

	X = D;

	// Now we perform the rounding process

	S = approx_SDP(H, X, nClasf + 1, (int) (*pruningRate * nClasf + .5), &vS);

	for (i = 0 ; i < 10 ; i++) {
		Stmp = approx_SDP(H, X, nClasf + 1, (int) (*pruningRate * nClasf + .5), &vStmp);

		if (vS > vStmp) {
			vS = vStmp;
			free(S);
			S = Stmp;
		} else {
			free(Stmp);
		}	
	}

	nsel = 0;
	for (i = 0 ; i < nClasf ; i++)
		if (S[ i ] != 0) {
			indexes[ i ] = 1;
			nsel++;
		} else {
			indexes[ i ] = 0;
		}

	free(S);

	printf("Numero de Clasificadores en conjunto final: %d. Score: %f\n", nsel, vS);

	// We free all the memory 

	for (i = 0; i < (nClasf + 1); i++) {
		free(H[ i ]); free(D[ i ]);
	}

	free(D); free(H);
}

int solve_SDP(double **H, double **D, int dim, int k) {

// The problem and solution data

struct blockmatrix C;
double *b;
struct constraintmatrix *constraints;

// Storage for the initial and final solutions.

struct blockmatrix X,Z;
double *y;
double pobj,dobj;

// blockptr will be used to point to blocks in constraint matrices.

struct sparseblock *blockptr;

// A return code for the call to easy_sdp().

int ret, i, j;

  /*
   * The first major task is to setup the C matrix and right hand side b.
   */

 /*
   * First, allocate storage for the C matrix.  We have three blocks, but
   * because C starts arrays with index 0, we have to allocate space for
   * four blocks- we'll waste the 0th block.  Notice that we check to
   * make sure that the malloc succeeded.
   */

	C.nblocks=1;
	C.blocks=(struct blockrec *)malloc(2 * sizeof(struct blockrec));
	if (C.blocks == NULL) {
		printf("Couldn't allocate storage for C!\n");
		return -1;
	}

  /*
   * Setup the first block.
   */

	C.blocks[1].blockcategory=MATRIX;
	C.blocks[1].blocksize=dim;
	C.blocks[1].data.mat=(double *)malloc(dim * dim * sizeof(double));
	if (C.blocks[1].data.mat == NULL) {
		printf("Couldn't allocate storage for C!\n");
		return -1;
	}


  /*
   * Put the entries into the first block. We exchange signs because we are minimizing
   */
	
	for (i = 0; i < dim; i++)
		for (j = 0; j < dim; j++)
			C.blocks[1].data.mat[ ijtok(i + 1 ,j + 1, dim) ]= -H[ i ][ j ];

  /*
   * Allocate storage for the right hand side, b.
   */

	b = (double *) malloc((dim + 2) * sizeof(double));
	if (b==NULL) {
		printf("Failed to allocate storage for a!\n");
		return -1;
	}

	/*
	* Fill in the entries in b
	*/

	for (j = 2; j < dim + 2; j++) b[ j ] = 1;
	b[ 1 ] = 4 * k;

  /*
   * The next major step is to setup the two constraint matrices A1 ... Adim.
   * Again, because C indexing starts with 0, we have to allocate space for
   * one more constraint.  constraints[0] is not used.
   */

	constraints=(struct constraintmatrix *)malloc((dim+1+1)*sizeof(struct constraintmatrix));
	if (constraints==NULL) {
		printf("Failed to allocate storage for constraints!\n");
		return -1;
	}

  /*
   * Setup the A1 matrix.  Note that we start with block 1 of A1 and then
   * do block 1 of A1.  We do this in this order because the blocks will
   * be inserted into the linked list of A1 blocks in reverse order.
   */

  /*
   * Terminate the linked list with a NULL pointer.
   */

	constraints[1].blocks=NULL;

   // We process each constrain. The first one is tr(HX) = b[ 1 ]

	/*
	 * Allocate space for block 1.
	 */

	blockptr=(struct sparseblock *)malloc(sizeof(struct sparseblock));
	if (blockptr==NULL) {
		printf("Allocation of constraint block failed!\n");
		return -1;
	}

	/*
	 * Initialize block 1.
	 */

	blockptr->blocknum=1;
	blockptr->blocksize=dim;
	blockptr->constraintnum=1;
	blockptr->next=NULL;
	blockptr->nextbyblock=NULL;
	blockptr->entries=(double *) malloc((dim * (dim - 1) / 2 + dim + 1)*sizeof(double));

	if (blockptr->entries==NULL) {
		printf("Allocation of constraint block failed!\n");
		return -1;
	}

	blockptr->iindices=(int *) malloc((dim * (dim - 1) / 2 + dim + 1)*sizeof(int));
	if (blockptr->iindices==NULL) {
		printf("Allocation of constraint block failed!\n");
		return -1;
	}

	blockptr->jindices=(int *) malloc((dim * (dim - 1) / 2 + dim + 1)*sizeof(int));
	if (blockptr->jindices==NULL) {
		printf("Allocation of constraint block failed!\n");
		return -1;
	}
	
	/*
	 * We have  dim * (dim - 1) / 2 nonzero entries in the upper triangle of block 1 of A0.
	 */

	blockptr->numentries= dim * (dim - 1) / 2 + dim;

	/*
	 * We set the entries of the block that are given by D
	 */

	int n = 1;
	for (i = 0 ; i < dim ; i++)
		for (j = i ; j < dim ; j++) {
			blockptr->iindices[ n ]= i + 1;
			blockptr->jindices[ n ]= j + 1;
			blockptr->entries[ n ]= D[ i ][ j ];
			n++;
		}

	/*
	 * Insert block 1 into the linked list of A1 blocks.
	 */

	blockptr->next=constraints[ 1 ].blocks;
	constraints[ 1 ].blocks=blockptr;

	// We set the constrains diag(V) = e

	for (i = 1 ; i < dim  + 1; i++) {

		constraints[ i + 1 ].blocks=NULL;

		/*
		 * Allocate space for block 1
		 */
		
		blockptr=(struct sparseblock *)malloc(sizeof(struct sparseblock));
		if (blockptr==NULL) {
			printf("Allocation of constraint block failed!\n");
			return -1;
		}

		/*
		 * Initialize block 1.
		 */

		blockptr->blocknum=1;
		blockptr->blocksize=dim;
		blockptr->constraintnum=i + 1;
		blockptr->next=NULL;
		blockptr->nextbyblock=NULL;
		blockptr->entries=(double *) malloc((1+1)*sizeof(double));

		if (blockptr->entries==NULL) {
			printf("Allocation of constraint block failed!\n");
			return -1;
		}

		blockptr->iindices=(int *) malloc((1+1)*sizeof(int));
		if (blockptr->iindices==NULL) {
			printf("Allocation of constraint block failed!\n");
			return -1;
		}

		blockptr->jindices=(int *) malloc((1+1)*sizeof(int));
		if (blockptr->jindices==NULL) {
			printf("Allocation of constraint block failed!\n");
			return -1;
		}
		
		/*
		 * We have 1 nonzero entry in the upper triangle of block 3 of A1.
		 */

		blockptr->numentries=1;

		/*
		 * The entry in the 1,1 position of block 1 of A1 to Adim is 1.0
		 */

		blockptr->iindices[1]=i;
		blockptr->jindices[1]=i;
		blockptr->entries[1]=1.0;

		/*
		 * Note that the entry in the 2,2 position of block 3 of A1 is 0,
		 * So we don't store anything for it.
		 */

		/*
		 * Insert block 1 into the linked list of A1 blocks.
		 */

		blockptr->next=constraints[ i + 1 ].blocks;
		constraints[ i + 1 ].blocks=blockptr;
	}

	/*
	 * Write the problem out in SDPA sparse format.
	 */

	write_prob("/tmp/prob.dat-s", dim, dim + 1, C, b, constraints);

	/*
	 * Create an initial solution.  This allocates space for X, y, and Z,
	 * and sets initial values.
	 */

	initsoln(dim, dim + 1,C,b,constraints,&X,&y,&Z);

	/*
	 * Solve the problem.
	 */
 
	ret = easy_sdp(dim, dim + 1, C, b, constraints, 0.0, &X, &y, &Z, &pobj, &dobj);

	if (ret == 0)
		printf("The objective value is %.7e \n",(dobj + pobj) / 2);
	else
		printf("SDP failed.\n");

	/*
	 * Write out the problem solution.
	 */

	write_sol("/tmp/prob.sol",dim , dim + 1, X, y, Z);

	// We compute the chol decompositon of X
	
	chol(X);

	// We set H to the solution X

	for (i = 0; i < dim; i++)
		for (j = 0; j < dim; j++)
			D[ i ][ j ] = X.blocks[1].data.mat[ ijtok(i + 1 ,j + 1, dim) ];

	/*
	 * Free storage allocated for the problem and return.
	 */

	free_prob(dim, dim + 1, C, b, constraints, X, y, Z);

	return 1;
}


// Function that solves the SDP optimization problem by using the CSPD library

int *approx_SDP(double **H, double **X, int dim, int k, double *value) {

double unif1, unif2, minValue, sum;
int *S = (int*) malloc(sizeof(int) * (dim - 1));
double *U = (double*) malloc(sizeof(double) * dim); 
int *x = (int*) malloc(sizeof(int) * dim);
int min, i, j, k_tmp, l;

	// We generate a gaussian random variable N(0, X)
	
	for (i = 0 ; i < dim; i++) {
		unif1 = 1.0 * rand () / (RAND_MAX + 1.0);
		unif2 = 1.0 * rand () / (RAND_MAX + 1.0);
		U[ i ] = cos(2 * 3.1415 * unif1) * sqrt(2 * log(1 / unif2));
	}


	for (i = 0 ; i < dim; i++) {
		sum = 0;
		for (j = 0 ; j < dim ; j++) 
			sum += X[ j ][ i ] * U[ j ];
		x[ i ] = sum >= 0 ? 1 : -1;
	}

	//  We set to one those values equal to x[ 0 ]

	for (i = 1 ; i < dim; i++)
		if (x[ i ] == x[ 0 ])
			S[ i - 1 ] = 1;
		else 
			S[ i - 1 ] = 0;

	k_tmp = 0;
	for (i = 0 ; i < dim - 1 ; i++) 
		if (S[ i ] != 0) k_tmp ++;

	printf("Ensemble size before geedy algorithm: %d. Target size: %d.\n", k_tmp, k);

	// If k_tmp == k we finish

	if (k_tmp == k) {
		free(U);
		free(x);

		sum = 0;
		for (i = 0 ; i < dim - 1 ; i++)
			if (S[ i ] != 0)
				for (j = 0 ; j < dim -1 ; j++)
					sum += S[ j ] * H[ j + 1 ][ i + 1 ];
		*value = sum;
		return S;
	}

	// Now we complete the k terms by a greedy search procedure

	do {
		k_tmp = 0;
		for (i = 0 ; i < dim - 1 ; i++) 
			if (S[ i ] != 0) k_tmp ++;

		if (k_tmp == k) break;

		min = 0; 
		minValue = 1e300;

		// We check wherther we have to remove or add a classifier

		for (l = 0 ; l < dim -1 ; l++) {

			// We set a new S vector
			
			if (k_tmp < k) {
				if (S[ l ] == 1)
					continue;
				else
					S[ l ] = 1;
			} else
				if (S[ l ] == 0)
					continue;
				else
					S[ l ] = 0;

			// We check the viability of the new S

			sum = 0;
			for (i = 0 ; i < dim - 1 ; i++)
				if (S[ i ] != 0)
					for (j = 0 ; j < dim -1 ; j++)
						sum += S[ j ] * H[ j + 1 ][ i + 1 ];

			// We restore the S vector and update the best candidate

			if (k_tmp < k)
				S[ l ] = 0;
			 else 
				S[ l ] = 1;

			if (minValue > sum)  {
				min = l;
				minValue = sum;
				*value = sum;
			}
		}

		if (k_tmp < k) 
			S[ min ] = 1;
		else 
			S[ min ] = 0;

	} while (k_tmp != k);

	k_tmp = 0;
	for (i = 0 ; i < dim - 1 ; i++) 
		if (S[ i ] != 0) k_tmp ++;

	printf("Reached size: %d.\n", k_tmp);

	free(U);
	free(x);
	return S;
}

